import sys

import autograder.api.courses.assignments.submissions.fetch.course.attempts
import autograder.cli.courses.assignments.submissions.common

def run(arguments):
    result = autograder.api.courses.assignments.submissions.fetch.course.attempts.send(arguments,
            exit_on_error = True)

    count = 0
    for (email, grading_result) in result['grading-results'].items():
        if (grading_result is None):
            continue

        autograder.cli.courses.assignments.submissions.common.output_grading_result(grading_result,
                arguments.out_dir)
        count += 1

    print("Wrote %d submissions to '%s'." % (count, arguments.out_dir))

    return 0

def main():
    return run(_get_parser().parse_args())

def _get_parser():
    parser = autograder.api.courses.assignments.submissions.fetch.course.attempts._get_parser()

    parser.add_argument('-o', '--out-dir', dest = 'out_dir',
        action = 'store', type = str, default = '.',
        help = ('Where to create a new directory that contains the submission information.'
            + ' Any existing subdirectories will be removed.'
            + ' Defaults to the current directory.'))

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import sys

import autograder.api.courses.assignments.submissions.fetch.user.history
import autograder.submission

def run(arguments):
    result = autograder.api.courses.assignments.submissions.fetch.user.history.send(arguments,
            exit_on_error = True)

    if (not result['found-user']):
        print("No matching user found.")
        return 1

    history = result['history']

    if (not arguments.table):
        print("Found %d submissions." % (len(history)))
        for entry in history:
            print("    " + str(autograder.submission.SubmissionSummary.from_dict(entry)))

        return 0

    header = [
        'id',
        'score',
        'max_points',
        'grading_start_time',
        'message',
    ]

    print("\t".join(header))
    for entry in history:
        entry = autograder.submission.SubmissionSummary.from_dict(entry)
        entry = entry.to_dict()
        print("\t".join([str(entry[key]) for key in header]))

    return 0

def main():
    return run(_get_parser().parse_args())

def _get_parser():
    parser = autograder.api.courses.assignments.submissions.fetch.user.history._get_parser()

    parser.add_argument('--table', dest = 'table',
        action = 'store_true', default = False,
        help = 'Output the results as a TSV table with a header (default: %(default)s).')

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



BASE_USER_HEADERS = ['email', 'name', 'role']

COURSE_USER_HEADERS = BASE_USER_HEADERS + ['lms-id']
SYNC_HEADERS = COURSE_USER_HEADERS + ['operation']

SERVER_USER_HEADERS = BASE_USER_HEADERS + ['courses']
COURSE_INFO_HEADERS = ['id', 'name', 'role']

INDENT = '    '
COURSE_USER_INFO_TYPE = 'course'
SERVER_USER_INFO_TYPE = 'server'

INDENT = '    '

SYNC_USERS_KEYS = [
    ('add-users', 'Added', 'add'),
    ('mod-users', 'Modified', 'mod'),
    ('del-users', 'Deleted', 'delete'),
    ('skip-users', 'Skipped', 'skip'),
]

USER_OP_KEYS = [
    ('added', 'Added'),
    ('modified', 'Modified'),
    ('removed', 'Removed'),
    ('skipped', 'Skipped'),
    ('not-exists', 'Not Exists'),
    ('emailed', 'Emailed'),
    ('enrolled', 'Enrolled'),
    ('dopped', 'Dropped'),
]

USER_OP_ERROR_KEYS = [
    ('validation-error', 'Validation Error'),
    ('system-error', 'System Error'),
    ('communication-error', 'Communication Error'),
]

ALL_USER_OP_KEYS = [
    ('email', 'Email'),
] + USER_OP_KEYS + USER_OP_ERROR_KEYS

# Set course_users to True if listing course users, False for server users.
# An error will be raised if a user of a different type is found.
def list_users(users, course_users, table = False, normalize = False):
    if (course_users):
        if (table):
            _list_users_table(users, True)
        else:
            _list_course_users(users)
    else:
        if (table):
            if (normalize):
                _list_server_users_table_normalize(users)
            else:
                _list_users_table(users, False)
        else:
            _list_server_users(users)

# Set course_users to True if listing course users, False for server users.
# An error will be raised if a user of a different type is found.
def _list_users_table(users, course_users, header = True, keys = []):
    expected_user_info_type = ''
    if (course_users):
        expected_user_info_type = COURSE_USER_INFO_TYPE
        if (len(keys) == 0):
            keys = COURSE_USER_HEADERS
    else:
        expected_user_info_type = SERVER_USER_INFO_TYPE
        if (len(keys) == 0):
            keys = SERVER_USER_HEADERS

    rows = []
    for user in users:
        if (user['type'] != expected_user_info_type):
            raise ValueError("Invalid user type for listing users table: Expected: '%s',"
                    + " actual: '%s'.", expected_user_info_type, user['type'])

        row = [user[key] for key in keys]
        rows.append(row)

    _print_tsv(rows, header, keys)

def _list_course_users(users, indent = ''):
    for i in range(len(users)):
        user = users[i]

        if (user['type'] != COURSE_USER_INFO_TYPE):
            raise ValueError("Invalid user type for listing course users: Expected: '%s',"
                    + " actual: '%s'.", COURSE_USER_INFO_TYPE, user['type'])

        if (i != 0):
            print()

        print(indent + "Email: " + user['email'])
        print(indent + "Name: " + user['name'])
        print(indent + "Role: " + user['role'])
        print(indent + "LMS ID: " + user['lms-id'])

def _list_server_users(users, indent = ''):
    for i in range(len(users)):
        user = users[i]

        if (user['type'] != SERVER_USER_INFO_TYPE):
            raise ValueError("Invalid user type for listing server users: Expected: '%s',"
                    + "actual: '%s'.", SERVER_USER_INFO_TYPE, user['type'])

        if (i != 0):
            print()

        print(indent + "Email: " + user['email'])
        print(indent + "Name: " + user['name'])
        print(indent + "Role: " + user['role'])
        print(indent + "Courses:")

        for j, course in enumerate(user['courses']):
            if (j != 0):
                print()

            print(indent + INDENT + "ID: " + user['courses'][course]['id'])
            print(indent + INDENT + "Name: " + user['courses'][course]['name'])
            print(indent + INDENT + "Role: " + user['courses'][course]['role'])

def _list_server_users_table_normalize(users, header = True, keys = BASE_USER_HEADERS):
    rows = []
    for user in users:
        if (user['type'] != SERVER_USER_INFO_TYPE):
            raise ValueError("Invalid user type for listing normalized server users table:"
                    + " Expected: '%s', actual: '%s'.", SERVER_USER_INFO_TYPE, user['type'])

        row = [user[key] for key in keys]
        if ((user.get('courses') is None) or (len(user['courses']) == 0)):
            row = row + ['' for key in COURSE_INFO_HEADERS]
            rows.append(row)
        else:
            for course in user['courses']:
                course_row = row + [user['courses'][course][key] for key in COURSE_INFO_HEADERS]
                rows.append(course_row)

    header_keys = keys + ["course-" + course_key for course_key in COURSE_INFO_HEADERS]
    _print_tsv(rows, header, header_keys)

def list_sync_users(sync_users, table = False):
    if (table):
        _list_sync_users_table(sync_users)
    else:
        _list_sync_users(sync_users)

def _list_sync_users(sync_users):
    count = (len(sync_users['add-users'])
            + len(sync_users['mod-users'])
            + len(sync_users['del-users']))
    print("Synced %d users." % (count))

    for (key, label, _) in SYNC_USERS_KEYS:
        users = sync_users[key]
        if (len(users) == 0):
            continue

        print("%s Users:" % (label))
        _list_course_users(users, indent = INDENT)

def _list_sync_users_table(sync_users):
    print("\t".join(SYNC_HEADERS))

    for (key, _, op) in SYNC_USERS_KEYS:
        users = sync_users[key]
        for user in users:
            user['operation'] = op

        _list_users_table(users, True, header = False, keys = SYNC_HEADERS)

def list_add_users(result, table = False):
    errors = result['errors']
    if ((errors is not None) and (len(errors) > 0)):
        print("Encountered %d errors." % (len(errors)))
        for error in errors:
            print("    Index: %d, Email: '%s', Message: '%s'." % (
                error['index'], error['email'], error['message']))

    list_sync_users(result, table = table)

def _list_user_op_responses(results):
    error_count = 0
    for result in results:
        print(result['email'])
        for op_key, label in USER_OP_KEYS:
            if (result.get(op_key, None) is not None):
                print(INDENT + label)
                if (isinstance(result[op_key], list)):
                    for value in result[op_key]:
                        print(INDENT + INDENT + value)

        for error_key, label in USER_OP_ERROR_KEYS:
            if (result.get(error_key, None) is not None):
                error_count += 1
                print(INDENT + label)
                print(INDENT + INDENT + result[error_key]['message'])

    print()
    print("Processed %d users. Encountered %d errors." % (len(results), error_count))

def _list_user_op_responses_table(results, header = True, keys = ALL_USER_OP_KEYS):
    rows = []
    for result in results:
        for error_key, _ in USER_OP_ERROR_KEYS:
            if (result.get(error_key, None) is not None):
                result[error_key] = result[error_key]['message']

        rows.append([result.get(key, '') for key, _ in keys])

    _print_tsv(rows, header, [header_key for _, header_key in keys])

def list_user_op_responses(results, table = False):
    if (table):
        _list_user_op_responses_table(results)
    else:
        _list_user_op_responses(results)

def _print_tsv(rows, header, header_keys):
    lines = []
    if (header):
        lines.append("\t".join(header_keys))

    for row in rows:
        lines.append("\t".join([str(value) for value in row]))

    print("\n".join(lines))



import os
import unittest

import autograder.util.submission

THIS_DIR = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))
DATA_DIR = os.path.join(THIS_DIR, "data")

class TestImport(unittest.TestCase):
    """
    Test the utilities for preparing submissions.
    """

    def test_simple(self):
        for ext in ['py', 'ipynb']:
            path = os.path.join(DATA_DIR, 'simple.' + ext)
            submission = autograder.util.submission.prepare(path)

            self.assertIn('__all__', dir(submission))
            self.assertIn('SOME_CONSTANT', dir(submission.__all__))
            self.assertEqual(submission.__all__.SOME_CONSTANT, 1)

            self.assertIn('simple', dir(submission))
            self.assertIn('SOME_CONSTANT', dir(submission.simple))
            self.assertEqual(submission.simple.SOME_CONSTANT, 1)

    def test_nested(self):
        for ext in ['py', 'ipynb']:
            path = os.path.join(DATA_DIR, 'nested')
            submission = autograder.util.submission.prepare(path)

            self.assertIn('__all__', dir(submission))
            self.assertIn('SOME_CONSTANT', dir(submission.__all__))
            self.assertEqual(submission.__all__.SOME_CONSTANT, 1)

            self.assertIn('nested1', dir(submission))
            self.assertIn('nested2', dir(submission.nested1))
            self.assertIn('nested', dir(submission.nested1.nested2))
            self.assertIn('SOME_CONSTANT', dir(submission.nested1.nested2.nested))
            self.assertEqual(submission.nested1.nested2.nested.SOME_CONSTANT, 1)



import autograder.api.common
import autograder.api.config

API_ENDPOINT = 'admin/logs/fetch'
API_PARAMS = [
    autograder.api.config.PARAM_COURSE_ID,
    autograder.api.config.PARAM_USER_EMAIL,
    autograder.api.config.PARAM_USER_PASS,

    autograder.api.config.APIParam('level',
            'The minimum level of log records to return (defaults to INFO).',
            required = False, parser_options = {'action': 'store', 'default': 'INFO'}),

    autograder.api.config.APIParam('after',
            'If supplied, only return log records after this datetime (as an RFC3339 string).',
            required = False),

    autograder.api.config.APIParam('past',
            ('If supplied, only return log records in this duration'
                + ' (using "h", "m", or "s" suffixes) (e.g., "24h", "10m", or "1h10m10s").'),
            required = False),

    autograder.api.config.APIParam('target-assignment',
            'If supplied, only return log records for this assignment.',
            required = False),

    autograder.api.config.APIParam('target-email',
            'If supplied, only return log records for this user.',
            required = False),
]

DESCRIPTION = 'Get log entries from the autograder server.'

def send(arguments, **kwargs):
    return autograder.api.common.handle_api_request(arguments, API_PARAMS, API_ENDPOINT, **kwargs)

def _get_parser():
    parser = autograder.api.config.get_argument_parser(
        description = DESCRIPTION,
        params = API_PARAMS)

    return parser



import sys

import autograder.api.courses.assignments.submissions.fetch.user.peek
import autograder.assignment

def run(arguments):
    result = autograder.api.courses.assignments.submissions.fetch.user.peek.send(arguments,
            exit_on_error = True)

    if (not result['found-user']):
        print("No matching user found.")
        return 1

    if (not result['found-submission']):
        print("No matching submission found.")
        return 2

    submission = autograder.assignment.GradedAssignment.from_dict(result['submission-result'])
    print(submission.report())
    return 0

def main():
    return run(_get_parser().parse_args())

def _get_parser():
    parser = autograder.api.courses.assignments.submissions.fetch.user.peek._get_parser()
    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import sys

import autograder.api.config
import autograder.api.users.upsert
import autograder.cli.common

def run(arguments):
    arguments = vars(arguments)

    password = arguments['new-pass']
    if (password != ''):
        password = autograder.util.hash.sha256_hex(password)

    arguments['raw-users'] = [{
        'email': arguments['new-email'],
        'name': arguments['new-name'],
        'role': arguments['new-role'],
        'pass': password,
        'course': arguments['new-course'],
        'course-role': arguments['new-course-role'],
        'course-lms-id': arguments['new-lms-id'],
    }]

    arguments['send-emails'] = not arguments['skip-emails']

    result = autograder.api.users.upsert.send(arguments, exit_on_error = True)

    autograder.cli.common.list_user_op_responses(result['results'], table = arguments['table'])
    return 0

def main():
    return run(_get_parser().parse_args())

def _get_parser():
    parser = autograder.api.users.upsert._get_parser()

    parser.add_argument('--skip-emails', dest = 'skip-emails',
        action = 'store_true', default = False,
        help = 'Skip sending any emails. Be aware that this may result in inaccessible'
        + ' information (default: %(default)s).')

    parser.add_argument('--new-email', dest = 'new-email',
        action = 'store', type = str, required = True,
        help = 'The email of the user to upsert.')

    parser.add_argument('--new-name', dest = 'new-name',
        action = 'store', type = str, default = '',
        help = 'The name of the user to upsert.')

    parser.add_argument('--new-role', dest = 'new-role',
        action = 'store', type = str, default = 'user',
        choices = autograder.api.constants.SERVER_ROLES,
        help = 'The role of the user to upsert (default: %(default)s).')

    parser.add_argument('--new-pass', dest = 'new-pass',
        action = 'store', type = str, default = '',
        help = 'The password of the user to upsert.'
            + ' If empty, the server will generate and email a password.')

    parser.add_argument('--new-course', dest = 'new-course',
        action = 'store', type = str, default = '',
        help = 'The course of the user to upsert.')

    parser.add_argument('--new-course-role', dest = 'new-course-role',
        action = 'store', type = str, default = 'student',
        choices = autograder.api.constants.COURSE_ROLES,
        help = 'The course role of the user to upsert (default: %(default)s).')

    parser.add_argument('--new-lms-id', dest = 'new-lms-id',
        action = 'store', type = str, default = '',
        help = 'The lms id of the user to upsert.')

    parser.add_argument('--table', dest = 'table',
        action = 'store_true', default = False,
        help = 'Output the results as a TSV table with a header (default: %(default)s).')

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import argparse
import sys
import traceback

import autograder.submission

DEFAULT_ASSIGNMENT = 'assignment.json'

def run(args):
    try:
        test_submissions = autograder.submission.fetch_test_submissions(args.submissions)
    except Exception as ex:
        print("Failed to load submission(s) from '%s': '%s'." % (args.submissions, ex))
        traceback.print_exc()
        return 101

    errors = 0
    for test_submission in test_submissions:
        success = False

        try:
            success = autograder.submission.run_test_submission(args.assignment,
                test_submission, args.debug)
        except Exception as ex:
            print("Failed to run submission '%s': '%s'." % (test_submission, ex))
            traceback.print_exc()

        if (not success):
            errors += 1

    print("Encountered %d error(s) while testing %d submissions." % (errors, len(test_submissions)))

    if (errors > 0):
        print("Faiure")
    else:
        print("Success")

    return errors

def _get_parser():
    parser = argparse.ArgumentParser(description =
        'Run a grader against multiple test assignments and ensure the output'
        + ' matches the expected output.')

    parser.add_argument('-a', '--assignment',
        action = 'store', type = str, required = False, default = DEFAULT_ASSIGNMENT,
        help = 'The path to a JSON file describing an assignment (default: %(default)s).')

    parser.add_argument('-s', '--submissions',
        action = 'store', type = str, required = True,
        help = 'The path to a dir containing one or more test submissions.')

    parser.add_argument('-d', '--debug', dest = 'debug',
        action = 'store_true', default = False,
        help = 'Enable additional output and leave behind grading artifacts (default: %(default)s)')

    return parser

def main():
    return run(_get_parser().parse_args())

if (__name__ == '__main__'):
    sys.exit(main())



"""
Filespecs (File Specifications) are objects in assignment configs that describe
files that need to exist.
These files can come from local files (path FileSpec), URLs (url FileSpec),
or git repos (git FileSpec).
"""

import os
import urllib.parse

import autograder.util.dirent
import autograder.util.git
import autograder.util.http

FILESPEC_TYPE_EMPTY = "empty"
FILESPEC_TYPE_NIL = "nil"
FILESPEC_TYPE_PATH = "path"
FILESPEC_TYPE_GIT = "git"
FILESPEC_TYPE_URL = "url"

def parse(data):
    """
    Parse and validate a FileSpec.
    Return a properly formatted FileSpec.

    This parsing is more lenient than the parsing done by the autograder server:
    https://github.com/eriq-augustine/autograder-server/blob/main/common/filespec.go#L37
    """

    if (data is None):
        return get_empty()

    if (isinstance(data, str)):
        if (data == ""):
            return get_empty()

        return get_path(data)

    if (not isinstance(data, dict)):
        raise FileSpecError("FileSpec is not the correct type (str ot dict): '%s' (%s)." % (
            data, type(data)))

    if ('type' not in data):
        raise FileSpecError("FileSpec is missing 'type' field: '%s'." % (data))

    data['type'] = data['type'].strip().lower()

    spec_type = data['type']
    if (spec_type == FILESPEC_TYPE_EMPTY):
        return get_empty()
    elif (spec_type == FILESPEC_TYPE_NIL):
        return get_nil()
    elif (spec_type == FILESPEC_TYPE_PATH):
        path = data.get('path', '').strip()
        if (path == ''):
            raise FileSpecError("Path FileSpec must have a non-empty path: '%s'." % (data))

        return get_path(path, dest = data.get('dest', ''))
    elif (spec_type == FILESPEC_TYPE_GIT):
        path = data.get('path', '').strip()
        if (path == ''):
            raise FileSpecError("Git FileSpec must have a non-empty path: '%s'." % (data))

        return get_git(path,
            dest = data.get('dest', ''),
            reference = data.get('reference', ''),
            username = data.get('username', ''),
            token = data.get('token', ''))
    elif (spec_type == FILESPEC_TYPE_URL):
        path = data.get('path', '').strip()
        if (path == ''):
            raise FileSpecError("URL FileSpec must have a non-empty path: '%s'." % (data))

        return get_url(path, dest = data.get('dest', ''))
    else:
        raise FileSpecError("FileSpec has unkown type ('%s'): '%s'." % (spec_type, data))

def get_empty():
    return {
        "type": FILESPEC_TYPE_EMPTY,
    }

def get_nil():
    return {
        "type": FILESPEC_TYPE_NIL,
    }

def get_path(path, dest = ''):
    if (dest == ''):
        dest = os.path.basename(path)

    return {
        "type": FILESPEC_TYPE_PATH,
        "path": path,
        "dest": dest,
    }

def get_git(path, dest = '', reference = '', username = '', token = ''):
    if (dest == ''):
        dest = os.path.splitext(os.path.basename(path))[0]

    if ((username != '') and (token == '')):
        raise FileSpecError(("If username is specified on a Git FileSpec,"
            + " then token must also be specified."))

    return {
        "type": FILESPEC_TYPE_GIT,
        "path": path,
        "dest": dest,
        "reference": reference,
        "username": username,
        "token": token,
    }

def get_url(path, dest = ''):
    if (dest == ''):
        url = urllib.parse.urlparse(path)
        dest = os.path.basename(url.path)

    return {
        "type": FILESPEC_TYPE_URL,
        "path": path,
        "dest": dest,
    }

def copy(filespec, base_dir, dest_dir, only_contents):
    spec_type = filespec['type']
    if (spec_type in [FILESPEC_TYPE_EMPTY, FILESPEC_TYPE_NIL]):
        # no-op.
        return
    elif (spec_type == FILESPEC_TYPE_PATH):
        _copy_path(filespec['path'], filespec['dest'], base_dir, dest_dir, only_contents)
    elif (spec_type == FILESPEC_TYPE_GIT):
        _copy_git(filespec['path'], filespec['dest'], dest_dir,
                reference = filespec['reference'],
                username = filespec['username'],
                token = filespec['token'])
    elif (spec_type == FILESPEC_TYPE_URL):
        _copy_url(filespec['path'], filespec['dest'], dest_dir)
    else:
        raise FileSpecError("FileSpec has unkown type ('%s'): '%s'." % (spec_type, filespec))

def _copy_path(path, dest, base_dir, dest_dir, only_contents):
    source_path = path
    if ((not os.path.isabs(source_path)) and (base_dir != '')):
        source_path = os.path.join(base_dir, path)

    if (only_contents):
        dest_path = dest_dir

        if (dest != ''):
            dest_path = os.path.join(dest_dir, dest)
    else:
        filename = dest
        if (filename == ''):
            filename = os.path.basename(path)

        dest_path = os.path.join(dest_dir, filename)

    if (only_contents):
        autograder.util.dirent.copy_contents(source_path, dest_path)
    else:
        autograder.util.dirent.copy(source_path, dest_path)

def _copy_git(path, dest, dest_dir, reference = '', username = '', token = ''):
    if (reference == ''):
        reference = None

    if (username == ''):
        username = None

    if (token == ''):
        token = None

    dest_path = os.path.join(dest_dir, dest)

    autograder.util.git.ensure_repo(path, dest_path, update = True,
        ref = reference, username = username, token = token)

def _copy_url(path, dest, dest_dir):
    dest_path = os.path.join(dest_dir, dest)
    autograder.util.http.get(path, dest_path)

class FileSpecError(ValueError):
    pass



import hashlib

DEFAULT_ENCODING = 'utf8'

def sha256_hex(payload, encoding = DEFAULT_ENCODING):
    digest = hashlib.new('sha256')
    digest.update(payload.encode(encoding))
    return digest.hexdigest()



import getpass
import sys

import autograder.api.config
import autograder.api.users.password.change

def run(arguments):
    arguments = vars(arguments)

    if (arguments.get(autograder.api.config.PARAM_NEW_PASS.key, None) is None):
        arguments[autograder.api.config.PARAM_NEW_PASS.key] = _get_pass(arguments['stdin'])

    result = autograder.api.users.password.change.send(arguments, exit_on_error = True)

    if (result['duplicate']):
        print("Your new password must be different from your previous password.")
        return 1

    if (not result['success']):
        print("Your password was not changed.")
        return 2

    print("You have successfully changed your password.")
    return 0

def _get_pass(stdin):
    # Get the password from stdout without any prompt or confirmation.
    if (stdin):
        new_pass = ''
        while (len(new_pass) == 0):
            new_pass = sys.stdin.readline().strip()
        return new_pass

    # Use a prompt and confirmation.
    new_pass = ''
    confirmation = ''

    while ((len(new_pass) == 0) or (new_pass != confirmation)):
        new_pass = ''
        while (len(new_pass) == 0):
            new_pass = getpass.getpass('Password: ').strip()

        confirmation = ''
        while (len(confirmation) == 0):
            confirmation = getpass.getpass('Repeat Password: ').strip()

        if (new_pass != confirmation):
            print("Passwords do not match, try again.")

    return new_pass

def main():
    return run(_get_parser().parse_args())

def _get_parser():
    parser = autograder.api.users.password.change._get_parser()

    addendum = ('If no password is supplied on the command line,'
        + ' then you will be prompted for the password and confirmation.'
        + ' Use `-s`/`--stdin` to suppress the prompt (and confirmation),'
        + ' and just read the password from stdin.'
        + ' Space will be stripped off of password read through stdin'
        + ' (with or without `--stdin`.'
        + ' Empty passwords are not allowed.')

    parser.description += "\n\n" + addendum

    parser.add_argument('-s', '--stdin', dest = 'stdin',
        action = 'store_true', default = False,
        help = 'Suppress normal output and read the password from stdin (default: %(default)s)')

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import autograder.api.common
import autograder.api.config

API_ENDPOINT = 'users/upsert'
API_PARAMS = [
    autograder.api.config.PARAM_USER_EMAIL,
    autograder.api.config.PARAM_USER_PASS,

    autograder.api.config.PARAM_DRY_RUN,

    autograder.api.config.APIParam('skip-inserts',
        'Skip inserts (default: False).',
        required = False,
        parser_options = {'action': 'store_true', 'default': False}),

    autograder.api.config.APIParam('skip-updates',
        'Skip updates (default: False).',
        required = False,
        parser_options = {'action': 'store_true', 'default': False}),

    autograder.api.config.APIParam('send-emails',
        'Send any emails.',
        required = True, cli_param = False),

    autograder.api.config.APIParam('raw-users',
        'A list of users to upsert.',
        required = True, cli_param = False),
]

DESCRIPTION = 'Upsert one or more users to the server (update if exists, insert otherwise).'

def send(arguments, **kwargs):
    return autograder.api.common.handle_api_request(arguments, API_PARAMS, API_ENDPOINT, **kwargs)

def _get_parser():
    parser = autograder.api.config.get_argument_parser(
        description = DESCRIPTION,
        params = API_PARAMS)

    return parser



import sys

import autograder.api.constants
import autograder.api.users.upsert
import autograder.cli.common
import autograder.util.hash

def run(arguments):
    arguments = vars(arguments)

    arguments['raw-users'] = _load_users(arguments['path'])
    arguments['send-emails'] = not arguments['skip-emails']

    result = autograder.api.users.upsert.send(arguments, exit_on_error = True)

    autograder.cli.common.list_user_op_responses(result['results'], table = arguments['table'])
    return 0

def _load_users(path):
    users = []

    with open(path, 'r') as file:
        lineno = 0
        for line in file:
            lineno += 1

            line = line.strip()
            if (line == ""):
                continue

            parts = line.split("\t")
            parts = [part.strip() for part in parts]

            if (len(parts) > 7):
                raise ValueError(
                    "File ('%s') line (%d) has too many values. Max is 7, found %d." % (
                        path, lineno, len(parts)))

            email = parts.pop(0)

            password = ''
            if (len(parts) > 0):
                password = parts.pop(0)
                if (password != ''):
                    password = autograder.util.hash.sha256_hex(password)

            name = ''
            if (len(parts) > 0):
                name = parts.pop(0)

            role = 'user'
            if (len(parts) > 0):
                role = parts.pop(0)
                role = role.lower()

            if (role not in autograder.api.constants.SERVER_ROLES):
                raise ValueError(
                    "File ('%s') line (%d) has an invalid role '%s'." % (
                        path, lineno, role))

            course = ''
            if (len(parts) > 0):
                course = parts.pop(0)

            course_role = 'unknown'
            if (len(parts) > 0):
                course_role = parts.pop(0)
                course_role = course_role.lower()

            if (course_role not in autograder.api.constants.COURSE_ROLES):
                raise ValueError(
                    "File ('%s') line (%d) has an invalid course role '%s'." % (
                        path, lineno, course_role))

            course_lms_id = ''
            if (len(parts) > 0):
                course_lms_id = parts.pop(0)

            users.append({
                'email': email,
                'pass': password,
                'name': name,
                'role': role,
                'course': course,
                'course-role': course_role,
                'course-lms-id': course_lms_id,
            })

    return users

def main():
    return run(_get_parser().parse_args())

def _get_parser():
    parser = autograder.api.users.upsert._get_parser()

    parser.description = ('Upsert users to the course from a TSV file.'
                + ' (Update if exists, otherwiese insert).')

    parser.add_argument('path', metavar = 'PATH',
        action = 'store', type = str,
        help = 'Path to a TSV file where each line contains up to seven columns:'
                + ' [email, pass, name, role, course, course-role, lms-id].'
                + ' Only the email is required. Leading and trailing whitespace is stripped'
                + ' from all fields, including pass. If pass is empty, a password will be'
                + ' randomly generated and emailed to the user.')

    parser.add_argument('--skip-emails', dest = 'skip-emails',
        action = 'store_true', default = False,
        help = 'Skip sending any emails. Be aware that this may result in inaccessible'
                + ' information (default: %(default)s).')

    parser.add_argument('--table', dest = 'table',
        action = 'store_true', default = False,
        help = 'Output the results as a TSV table with a header (default: %(default)s).')

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import json
import os
import sys

import requests

import autograder.api.config
import autograder.api.constants
import autograder.api.error
import autograder.util.timestamp

def handle_api_request(arguments, params, endpoint, exit_on_error = False, files = []):
    """
    Given arguments (usually from argparse), API params, and an endpoint,
    make an API request.
    Will raise an error on any error or error response.
    """

    try:
        return _handle_api_request(arguments, params, endpoint, exit_on_error, files)
    except autograder.api.error.AutograderError as ex:
        if (exit_on_error):
            print("ERROR: " + ex.args[0], file = sys.stderr)
            autograder.api.error.exit_from_error(1)

        raise ex

def _handle_api_request(arguments, params, endpoint, exit_on_error, files):
    config = autograder.api.config.get_tiered_config(arguments)
    data, extra = autograder.api.config.parse_api_config(config, params,
            exit_on_error = exit_on_error)

    return send_api_request(endpoint, data = data, files = files, **extra)

def send_api_request(endpoint, server = None, verbose = False, data = {}, files = [], **kwargs):
    """
    Make an autograder API request.
    On a failure statue, an error will be raised with error information.
    Otherwise (including a successful response with a failed operation (soft error)),
    the content of the response will be returned.
    """

    if ((server is None) or (server == '')):
        raise autograder.api.error.APIError(None, "No server provided.")

    server = server.rstrip('/')
    endpoint = endpoint.lstrip('/')

    url = "%s/api/%s/%s" % (server, autograder.api.constants.API_VERSION, endpoint)

    post_files = {}
    for path in files:
        filename = os.path.basename(path)
        if (filename in post_files):
            raise autograder.api.error.APIError(None, "Cannot submit duplicate filenames ('%s')."
                % (filename))

        post_files[filename] = open(path, 'rb')

    if (verbose):
        print("\nAutograder Request Data:\n---\n%s\n---\n" % (json.dumps(data, indent = 4)))

    try:
        raw_response = requests.request(
            method = 'POST',
            url = url,
            data = {autograder.api.constants.API_REQUEST_JSON_KEY: json.dumps(data)},
            files = post_files)
    except requests.exceptions.ConnectionError:
        raise autograder.api.error.ConnectionError(("Could not connect to autograder server"
            + " '%s'." % (server)
            + " This is a networking issue"
            + " (e.g., network down, server down, wrong server address),"
            + " not an authentication issue."))

    for file in post_files.values():
        file.close()

    try:
        response = raw_response.json()
    except Exception as ex:
        raise autograder.api.error.APIError(None, "Autograder response does not contain valid JSON."
            + " Contact a server admin with the following. Response:\n---\n%s\n---" % (
                raw_response.text)) from ex

    if (verbose):
        print("\nAutograder Response:\n---\n%s\n---\n" % (json.dumps(response, indent = 4)))

    if (not response.get(autograder.api.constants.API_RESPONSE_KEY_SUCCESS, False)):
        message = 'Request to the autograder failed.'
        if (autograder.api.constants.API_RESPONSE_KEY_MESSAGE in response):
            message = ("Failed to complete operation: %s" %
                response[autograder.api.constants.API_RESPONSE_KEY_MESSAGE])

            # Replace any timestamps in the message.
            message = autograder.util.timestamp.convert_message(message, pretty = True)

        code = response.get("status", None)
        raise autograder.api.error.APIError(code, message)

    return response[autograder.api.constants.API_RESPONSE_KEY_CONTENT]



import autograder.api.common
import autograder.api.config

API_ENDPOINT = 'lms/upload/scores'
API_PARAMS = [
    autograder.api.config.PARAM_COURSE_ID,
    autograder.api.config.PARAM_USER_EMAIL,
    autograder.api.config.PARAM_USER_PASS,

    autograder.api.config.APIParam('assignment-lms-id',
        'The LMS ID of the assignment to upload scores to.',
        required = True),

    autograder.api.config.APIParam('scores',
        'A list of scores to upload.',
        required = True, cli_param = False),
]

DESCRIPTION = ("Upload scores from a tab-separated file to the course's LMS."
        + " The file should not have headers, and should have two columns: email and score.")

def send(arguments, **kwargs):
    return autograder.api.common.handle_api_request(arguments, API_PARAMS, API_ENDPOINT, **kwargs)

def _get_parser():
    parser = autograder.api.config.get_argument_parser(
        description = DESCRIPTION,
        params = API_PARAMS)

    return parser



import json
import sys

import autograder.api.admin.fetchlogs
import autograder.util.timestamp

LEVEL_TO_STRING = {
    -20: 'TRACE',
    -10: 'DEBUG',
    0: 'INFO',
    10: 'WARN',
    20: 'ERROR',
    30: 'FATAL',
    100: 'OFF',
}

def run(arguments):
    result = autograder.api.admin.fetchlogs.send(arguments, exit_on_error = True)

    if (not result['success']):
        print("Error fetching logs:")
        for message in result['error-messages']:
            print("    " + message)

        return 1

    if (arguments.json):
        print(json.dumps(result['results'], indent = 4))
    else:
        for record in result['results']:
            print(_log_record_str(record))

    return 0

def _log_record_str(record):
    level = "Unknown (%d)" % (record['level'])
    if (record['level'] in LEVEL_TO_STRING):
        level = LEVEL_TO_STRING[record['level']]

    timestamp = autograder.util.timestamp.get(record['timestamp'], pretty = True)

    message = record['message']

    attributes = record.get('attributes', {})

    for key in ['course', 'assignment', 'user']:
        if (key in record):
            attributes[key] = record[key]

    if (len(attributes) > 0):
        message += (" | " + json.dumps(attributes))

    return "%s [%5s] %s" % (timestamp, level, message)

def main():
    return run(_get_parser().parse_args())

def _get_parser():
    parser = autograder.api.admin.fetchlogs._get_parser()

    parser.add_argument('--json', dest = 'json',
        action = 'store_true', default = False,
        help = ('Output the results as a JSON array instead of a more human-readable format'
            + ' (default: %(default)s).'))

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import sys

# Control if exit_from_error() should actually exit.
# Testing infrastructure can set this to control exit behavior.
_exit_on_error_for_testing = True

def exit_from_error(exit_status = 1):
    """
    Exit because an error occurred.
    Tetsing infrastructure can set _exit_on_error_for_testing to false to avoid exiting.
    """

    if (not _exit_on_error_for_testing):
        return

    sys.exit(exit_status)

class AutograderError(Exception):
    pass

class APIError(AutograderError):
    def __init__(self, code, message):
        super().__init__(message)
        self.code = code

class ConnectionError(AutograderError):
    pass



import sys

import autograder.api.config
import autograder.api.users.list
import autograder.cli.common

def run(arguments):
    result = autograder.api.users.list.send(arguments, exit_on_error = True)
    autograder.cli.common.list_users(result['users'], False, table = arguments.table,
        normalize = arguments.normalize)
    return 0

def main():
    return run(_get_parser().parse_args())

def _get_parser():
    parser = autograder.api.users.list._get_parser()

    parser.add_argument('--table', dest = 'table',
        action = 'store_true', default = False,
        help = 'Output the results as a TSV table with a header (default: %(default)s).')

    parser.add_argument('--normalize', dest = 'normalize',
        action = 'store_true', default = False,
        help = 'Normalize the TSV table to include at most one course enrollment per line, users'
            + ' enrolled in multiple courses will appear multiple times. Each line contains the'
            + ' following columns: [email, name, role, course-id, course-name, course-role]. If'
            + ' a user is not enrolled in any courses, they will appear once with empty course'
            + ' information. Only applies if --table is set (default: %(default)s).')

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import multiprocessing
import sys
import time
import traceback

REAP_TIME_SEC = 5

# Return: (success, function return value)
# On timeout, success will be false and the value will be None.
# On error, success will be false and value will be the string stacktrace.
# On successful completion, success will be true and value may be None (if nothing was returned).
def with_timeout(timeout, function):
    if (not sys.platform.startswith('linux')):
        # Mac and Windows have some pickling issues with multiprocessing.
        # Just run them without a timeout.
        # Any autograder will be run on a Linux machine and will be safe.
        start_time = time.time()
        value = function()
        runtime = time.time() - start_time

        if (runtime > timeout):
            return (False, None)

        return (True, value)

    result = multiprocessing.Queue(1)

    # Note that we use processes instead of threads so they can be more completely killed.
    process = multiprocessing.Process(target = _invoke_helper, args = (result, function))
    process.start()

    # Wait for at most the timeout.
    process.join(timeout)

    # Check to see if the process is still running.
    if (process.is_alive()):
        # Kill the long-running process.
        process.terminate()

        # Try to reap the process once before just giving up on it.
        process.join(REAP_TIME_SEC)

        return (False, None)

    # Check to see if the process explicitly existed (like via sys.exit()).
    if (result.empty()):
        return (False, 'Code explicitly exited (like via sys.exit()).')

    value, error = result.get()

    if (error is not None):
        exception, stacktrace = error
        return (False, stacktrace)

    return (True, value)

def _invoke_helper(result, function):
    value = None
    error = None

    try:
        value = function()
    except Exception as ex:
        error = (ex, traceback.format_exc())

    sys.stdout.flush()

    result.put((value, error))
    result.close()



import sys

import autograder.api.users.password.reset

def run(arguments):
    autograder.api.users.password.reset.send(arguments, exit_on_error = True)
    print("The server has successfully processed your request.")
    print("Check your email for the new password.")
    return 0

def main():
    return run(_get_parser().parse_args())

def _get_parser():
    parser = autograder.api.users.password.reset._get_parser()
    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import os
import unittest

import autograder.code

THIS_DIR = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))
DATA_DIR = os.path.join(THIS_DIR, "data")

class TestImport(unittest.TestCase):
    """
    Test the utilities for importing code.
    """

    def test_extract_base(self):
        for ext in ['py', 'ipynb']:
            path = os.path.join(DATA_DIR, 'simple.' + ext)
            source_code = autograder.code.extract_code(path)
            self.assertEqual(source_code, "SOME_CONSTANT = 1")

    def test_import_base(self):
        for ext in ['py', 'ipynb']:
            # The code should be executed, causing some_int to become -1.
            path = os.path.join(DATA_DIR, 'base.' + ext)
            module = autograder.code.import_path(path)

            self.assertEqual(module.SOME_CONSTANT, 1)
            self.assertEqual(module.some_int, -1)

            path = os.path.join(DATA_DIR, 'base_with_raise.' + ext)
            try:
                module = autograder.code.import_path(path)
                self.fail("Import with raise did not fail as expected.")
            except RuntimeError:
                # Expected.
                pass

    def test_sanitize_import_base(self):
        for ext in ['py', 'ipynb']:
            for basename in ['base', 'base_with_raise']:
                path = os.path.join(DATA_DIR, basename + '.' + ext)
                module = autograder.code.sanitize_and_import_path(path)

                self.assertEqual(module.SOME_CONSTANT, 1)
                self.assertIn('random', dir(module))
                self.assertNotIn('some_int', dir(module))
                self.assertIn('some_function', dir(module))
